﻿#include "pch.h"
#include "UnderlineDemoRenderer.h"
#include "Common/DirectXHelper.h"
#include "UnderlineSpecifier.h"

using namespace UnderlineDemo;

using namespace D2D1;
using namespace Microsoft::WRL;

UnderlineDemoRenderer::UnderlineDemoRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) :
m_deviceResources(deviceResources)
{
    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    ComPtr<IDWriteTextFormat> textFormat;

    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Times New Roman",
            nullptr,
            DWRITE_FONT_WEIGHT_NORMAL,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            24.0f,
            L"en-us",
            &textFormat)
        );

    m_text = L"This paragraph of text rendered with "
             L"DirectWrite is based on "
             L"IDWriteTextFormat and IDWriteTextLayout "
             L"objects and is capable of different "
             L"RBG RGB foreground colors, such as red, "
             L"green, and blue by passing brushes to "
             L"the SetDrawingEffect method. "
             L"Or, an instance of an UnderlineFormatter "
             L"can be passed to that method for double "
             L"underline, triple underline, double "
             L"strikethrough, triple strikethrough, "
             L"or combinations thereof.";

    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextLayout(
            m_text.c_str(),
            (uint32) m_text.length(),
            textFormat.Get(),
            440.0f,         // Max width for paragraph
            std::numeric_limits<float>::infinity(),
            &m_textLayout)
        );

    // Set character formatting
    DWRITE_TEXT_RANGE textRange;

    std::wstring strFind = L"IDWriteTextFormat";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    strFind = L"IDWriteTextLayout";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    strFind = L"SetDrawingEffect";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    strFind = L"UnderlineFormatter";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    // Set custom underlining and strikethrough
    strFind = L"double underline";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(2, 0), textRange)
        );

    strFind = L"triple underline";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(3, 0), textRange)
        );

    strFind = L"double strikethrough";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(0, 2), textRange)
        );

    strFind = L"triple strikethrough";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(0, 3), textRange)
        );

    strFind = L"combinations";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(3, 2), textRange)
        );

    strFind = L"thereof";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();

    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(new UnderlineSpecifier(2, 3), textRange)
        );

    // Get text metrics
    DX::ThrowIfFailed(
        m_textLayout->GetMetrics(&m_textMetrics)
        );

    // Instantiate UnderlineFormatter
    m_underlineFormatter = new UnderlineFormatter();

    CreateDeviceDependentResources();
}

void UnderlineDemoRenderer::CreateDeviceDependentResources()
{
    // Create brushes and set them
    ID2D1DeviceContext1* context = m_deviceResources->GetD2DDeviceContext();

    ComPtr<ID2D1SolidColorBrush> redBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Red), &redBrush)
        );

    ComPtr<ID2D1SolidColorBrush> greenBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Green), &greenBrush)
        );

    ComPtr<ID2D1SolidColorBrush> blueBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Blue), &blueBrush)
        );

    DWRITE_TEXT_RANGE textRange;

    std::wstring strFind = L"RBG";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    textRange.length = 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    strFind = L"RGB";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    textRange.length = 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    strFind = L" red";      // avoid "rendered"
    textRange.startPosition = m_text.find(strFind.data()) + 1;
    textRange.length = strFind.length() - 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    strFind = L"green";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    strFind = L"blue";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    // Default text brush
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}

void UnderlineDemoRenderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void UnderlineDemoRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void UnderlineDemoRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::AliceBlue));

    // Center text on the screen
    Matrix3x2F screenTranslation = Matrix3x2F::Translation(
        (logicalSize.Width - m_textLayout->GetMaxWidth()) / 2,
        (logicalSize.Height - m_textMetrics.height) / 2);

    context->SetTransform(screenTranslation *
        m_deviceResources->GetOrientationTransform2D());

    // Display paragraph of text with custom text renderer
    D2D1_POINT_2F origin = Point2F();
    DrawingContext drawingContext(context, m_blackBrush.Get());

    m_textLayout->Draw(&drawingContext, m_underlineFormatter.Get(),
                       origin.x, origin.y);

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

